#include "MRRegularMapMesher.h"
#include "MRPointsLoad.h"
#include "MRRegularGridMesh.h"

namespace MR
{

tl::expected<void, std::string> RegularMapMesher::loadSurfacePC( const std::filesystem::path& path )
{
    auto res = PointsLoad::fromAnySupportedFormat( path );
    if ( !res.has_value() )
        return tl::make_unexpected( res.error() );

    surfacePC_ = std::make_unique<PointCloud>( std::move( res.value() ) );
    return {};
}

void RegularMapMesher::setSurfacePC( const std::shared_ptr<PointCloud>& surfacePC )
{
    surfacePC_ = surfacePC;
}

tl::expected<void, std::string> RegularMapMesher::loadDirectionsPC( const std::filesystem::path& path )
{
    auto res = PointsLoad::fromAnySupportedFormat( path );
    if ( !res.has_value() )
        return tl::make_unexpected( res.error() );
    
    directionsPC_ = std::make_unique<PointCloud>( std::move( res.value() ) );
    return {};
}

void RegularMapMesher::setDirectionsPC( const std::shared_ptr<PointCloud>& directionsPC )
{
    directionsPC_ = directionsPC;
}

tl::expected<void, std::string> RegularMapMesher::loadDistances( int width, int height, const std::filesystem::path& path )
{
    width_ = width;
    height_ = height;
    std::error_code ec;
    if ( std::filesystem::file_size( path, ec ) != ( height_ * width_ * sizeof( float ) ) )
    {
        distances_.clear();
        return tl::make_unexpected( "Distances file size is not equal height * width * sizeof(float)" );
    }
    std::ifstream ifs( path, std::ios::binary );
    distances_.resize( width_ );
    ifs.read( (char*) distances_.data(), distances_.size() * sizeof( float ) );
    return {};
}

void RegularMapMesher::setDistances( int width, int height, const std::vector<float>& distances )
{
    width_ = width;
    height_ = height;
    distances_ = distances;
}

tl::expected<Mesh, std::string> RegularMapMesher::createMesh() const
{
    auto refSize = width_ * height_;
    if ( !surfacePC_ )
        return tl::make_unexpected( "Surface Point Cloud is not loaded" );
    if ( surfacePC_->points.size() != refSize )
        return tl::make_unexpected( "Surface Point Cloud size is not equal width*height" );

    if ( !directionsPC_ )
        return tl::make_unexpected( "Directions Point Cloud is not loaded" );
    if ( directionsPC_->points.size() != width_ )
        return tl::make_unexpected( "Directions Point Cloud size is not equal width" );

    if ( distances_.empty() )
        return tl::make_unexpected( "Distances file is not loaded" );
    if ( distances_.size() != refSize )
        return tl::make_unexpected( "Distances size is not equal width*height" );


    auto mesh = makeRegularGridMesh( width_, height_, [&]( size_t x, size_t y )
    {
        return distances_[x + y * width_] != 0.0;
    },
                                [&]( size_t x, size_t y )
    {
        VertId idx = VertId( x + y * width_ );
        Vector3f org = surfacePC_->points[idx];
        Vector3f dest = directionsPC_->points[VertId(x)];
        return org + ( dest - org ).normalized() * ( 1.0f / distances_[idx] );
    } );
    mesh.topology.flipOrientation();
    return mesh;
}

}